"""
This module contains all the necessary functionality for the authentication of a request to the pubtrack app.
"""
from typing import Dict, Any
from abc import ABC, abstractmethod


class AbstractAuthentication(ABC):
    """
    This is the abstract base class for any "Authentication" object.

    **Background**

    The pubtrack application is a web application, which is running a REST API. So data transfer to and from this
    application is managed by HTTP requests. But only registered users are supposed to be able to make successful
    requests. Thus the need for an authentication. Roughly speaking that is something you can add to the request to
    ensure, that you are indeed a registered user.

    **Details**

    Each subclass of this base class is supposed to represent and implement the functionality to add such a
    authentication to a request! For that this class is designed to be a callable. The input parameters for this
    call are the the keyword arguments for the http request and the return will also be dict with mostly the same
    keyword arguments, except with additional authentication arguments added on top. To actually implement this
    the child classes will have to implement the "update_request_kwargs" abstract method.

    **Example**

    The following example illustrates how a Authentication object is used theoretically.

    .. code-block::python

        # "Authentication" is a sub class of "AbstractAuthentication" and implements
        # the "update_request_kwargs" method appropriately
        authenticate = Authentication(config)
        request_kwargs = {}
        authenticated_request_kwargs = authenticate(request_kwargs)

    """

    def __init__(self, config: Any):
        self.config = config

    def __call__(self, kwargs: dict) -> Dict[str, Any]:
        """
        This method will accept the kwargs dict of a http request towards an endpoint and then add additional arguments
        to it, which authenticate this request properly.

        :param kwargs: The keyword arguments for a http request to a pubtrack endpoint.
        :return: The authenticated keyword arguments dict for that request
        """
        return self.update_request_kwargs(kwargs)

    # TO BE IMPLEMENTED
    # -----------------

    @abstractmethod
    def update_request_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """
        This method has to be implemented for the subclass!

        It is supposed to add appropriate authentication details to the dict of keyword arguments.

        :param kwargs: The keyword arguments for a http request to a pubtrack endpoint.
        :return: The authenticated keyword arguments dict for that request
        """
        raise NotImplementedError


class TokenAuthentication(AbstractAuthentication):
    """
    Subclass of AbstractAuthentication, which implements the token authentication for the pubtrack application

    **Background**

    With token authentication, special "token" strings are generated by the web application and distributed to the
    users. The users then have to add this token string to each request to identify themselves.

    **Details**

    This class will read the token of the user from the given "config" object and add it as a keyword argument to the
    request
    """

    # IMPLEMENT "AbstractAuthentication"
    # ----------------------------------

    def update_request_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        token = self.config.get_token()
        headers = self.authentication_headers(token)

        kwargs.update({'headers': headers})
        return kwargs

    # HELPER METHODS
    # --------------

    @classmethod
    def authentication_headers(cls, token: str) -> Dict[str, str]:
        """
        This method will generate the necessary keyword key-value structure, which is needed for token authentication
        with the pubtrack application, given the actual "token" of the user.

        :param token: The token which is to be used for the authentication.
        :return:
        """
        return {
            'Authorization': 'TOKEN {}'.format(token)
        }



"""Console script for pypubtrack."""
import sys
import os
import click
import shutil

from pykitopen import KitOpen
from pykitopen.search import YearBatching
from pykitopen.publication import Publication
from pykitopen.config import DEFAULT

from pypubtrack.util import (get_config_path,
                             get_version,
                             check_installation,
                             get_installation_path,
                             init_installation,
                             author_name_kitopen)
from pypubtrack.config import Config
from pypubtrack.pypubtrack import Pubtrack


@click.group('pypubtrack', invoke_without_command=True)
@click.option('--version', '-v', is_flag=True, help='Print the currently installed version of the program')
@click.option('--config', '-c', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help='Provide an alternative config file for the project')
@click.pass_context
def cli(ctx, version, config):
    """
    Command line client interface to access "pubtrack" web services.
    """
    ctx.ensure_object(dict)

    # If the "version" option was passed, the user simply wants to display the version number of the project.
    if version:
        # "get_version" loads the version string from the version file and returns it.
        version = get_version()
        click.secho('PYPUBTRACK VERSION')
        click.secho(version, bold=True)
        return 0

    # Here we need to try and load the config file. BUT if the command which was invoked is the "init" command, then
    # we dont! Because the init command creates the config file first, logically it cannot already exist then.
    if ctx.invoked_subcommand != 'init':
        # The config singleton in created. It is then loaded with the config file which was passed as an option to
        # this command. If not option was given, it attempts to load the config file from the installation folder in
        # the users home directory.
        cfg = Config()
        if config:
            cfg.load_file(config)
        else:
            # We will print an error message, if the installation folder does not exists already
            if not check_installation():
                click.secho('No valid installation folder/config file exists on this machine yet!', fg='red')
                click.secho('HINT: Run the "init" command to create a new installation folder')
                return 1

            config_path = get_config_path()
            cfg.load_file(config_path)
        # Important: The config object is then saved into the context, which is being passed to the invoked sub commands
        ctx.obj['config'] = cfg


@click.command('init', short_help='Initializes the installation folder and the config file for this project')
@click.option('--force', '-f', is_flag=True, help='Forcefully deletes any current installation before creating new one')
@click.pass_context
def init(ctx, force):
    """
    Initializes an installation folder for this CLI program in the users HOME directory. This installation folder will
    contain a config file, which can be used to store access information such as the pubtrack URL and TOKEN for future
    uses of CLI commands.
    """
    folder_path = get_installation_path()
    # If "force" flag is true, then the user wants to delete the already existing installation first.
    if force:
        if os.path.exists(folder_path) and os.path.isdir(folder_path):
            # "rmtree" is necessary here to delete a whole folder structure with all it's contents recursively
            shutil.rmtree(folder_path)
            click.secho('Deleted previous installation folder', fg='green')
        else:
            click.secho('No previous installation folder existed.')

    # We are being nice here and check for a potential error. If the installation folder already exists we will warn
    # the user and print a hint message
    if os.path.exists(folder_path) and os.path.isdir(folder_path):
        click.secho('An installation folder already exists for this user!', fg='red')
        click.secho('HINT: run the "init" command with the "--force" flag to delete the previous one first')
        return 1

    # "init_installation" creates the installation folder and copies the config template into it.
    init_installation()
    click.secho('Created new installation folder at "{}"'.format(folder_path), fg='green', bold=True)


@click.command('update-scopus')
def update_scopus():
    click.secho('update scopus')


@click.command('update-kitopen')
@click.option('--verbose', '-v', is_flag=True, help='Print additional console output')
@click.option('--start', '-s', type=click.STRING, default='2015', help='The year to start the KITOpen query from')
@click.pass_context
def update_kitopen(ctx, verbose, start):
    """
    Update the pubtrack records with KITOpen information.

    This commend will fetch the names of all authors from the pubtrack application. These names will be used to
    query the KITOpen database. The data from KITOpen will then finally be used to update the publication records +
    of the pubtrack application.
    """
    config = ctx.obj['config']
    pubtrack = Pubtrack(config)

    # Getting the meta authors from the pubtrack site
    click.secho('Fetching author information from pubtrack.')
    author_names = []
    meta_authors = pubtrack.meta_author.get()['results']
    for meta_author in meta_authors:
        for author in meta_author['authors']:
            # "author_name_kitopen" returns a string with the authors name. This function essentially formats the name
            # in a way so that it can be used in a query string for the KITOpen database.
            name = author_name_kitopen(author['first_name'], author['last_name'])
            author_names.append(name)
            if verbose:
                click.secho(' * Adding author "{}" to query'.format(name))

    click.secho('==> Processing total of {} authors'.format(len(author_names)))

    # Setting up KITOpen API.
    # The "view" within the kitopen config defines how many details about a single publication record are supposed to
    # be returned as a result of the query. the batching strategy defines in what kind of batches the data is supposed
    # to be returned by the database. YearBatching is generally recommended to not hit the size limit.
    kitopen_config = DEFAULT.copy()
    kitopen_config['default_view'] = Publication.VIEWS.FULL
    kitopen_config['batching_strategy'] = YearBatching

    kitopen = KitOpen(kitopen_config)

    # Using the names of these authors to query kitopen results.
    # The "author" field of the search request is supposed to be the main query string. It will combine the names of
    # all the previously fetched authors with "or" directives between them.
    click.secho('Querying KITOpen database with previously fetched authors.')
    results = kitopen.search({
        'author': ' or '.join(author_names),
        'start': start,
        'end': ''
    })

    # Processing the kitopen results to update the pubtrack entries with that
    click.secho('Updating pubtrack records with KITOpen data')
    count_total = 0
    count_success = 0
    for publication in results:
        # The "publication" results of a KITOpen request save all their content in the internal "data" dict property.
        # Here we only consider publications, which have a valid DOI as that is the only way we can properly identify
        # them and associate them with the correct entries of the pubtrack application.
        if publication.data['doi']:
            try:
                # This is what I meant. The "get_by" function attempts to fetch a publication record from the pubtrack
                # app, which is uniquely identified by the doi from the KITOpen record.
                pub = pubtrack.publication.get_by(doi=publication.data['doi'])

                # Now we need to update this record of pubtrack. For that we are going to use a http PATCH request.
                # This way we only need to assemble a dict with the actually new data.
                # The important data from KITOpen are the ID and the POF structure
                patch = {
                    'on_kitopen': True,
                    'pof_structure': publication.data['pof_structure']
                }
                if not pub['kitopen_id']:
                    patch['kitopen_id'] = publication.data['id']

                # The patch request is identified with the UUID of the publication! This is why we had to get it before
                # in the first place. The uuid is the main identifier for the pubtrack app and is different from the
                # DOI.
                pubtrack.publication.patch(pub['uuid'], patch=patch)
                count_success += 1
                if verbose:
                    click.secho(' > updated publication {}'.format(publication.data['doi']))
            except Exception as e:
                if verbose:
                    click.secho(' ! Warning updating publication {}'.format(publication.data['doi']), fg='yellow')
            finally:
                count_total += 1

    click.secho('==> Updated {}/{} publications with KITOpen data'.format(count_success, count_total),
                fg='green', bold=True)

    return 0


cli.add_command(init)
cli.add_command(update_scopus)
cli.add_command(update_kitopen)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



"""Main module."""
from typing import Any, Dict, Iterable

from pypubtrack.config import Config
from pypubtrack.endpoint import Endpoint, AddEndpoint
from pypubtrack.util import exclude_keys


class PublicationsEndpoint(Endpoint):

    def get_endpoint(self):
        return 'publications'


class AuthorsEndpoint(Endpoint):

    def get_endpoint(self):
        return 'authors'


class AuthoringsEndpoint(Endpoint):

    def get_endpoint(self):
        return 'authorings'


class MetaAuthorsEndpoint(Endpoint):

    def get_endpoint(self):
        return 'meta-authors'


@AddEndpoint('publication', PublicationsEndpoint)
@AddEndpoint('author', AuthorsEndpoint)
@AddEndpoint('authoring', AuthoringsEndpoint)
@AddEndpoint('meta_author', MetaAuthorsEndpoint)
class Pubtrack:

    def __init__(self, config: Config):
        self.config = config
        self.url = self.config.get_url()

    # PUBLIC METHODS
    # --------------

    def import_publication(self, publication: Dict[str, Any]):
        base_publication = exclude_keys(publication, ['authors'])

        try:
            response = self.publication.post(base_publication)
            publication['uuid'] = response['uuid']
        except ConnectionError as err:
            print(err)

        self._import_publication_authors(publication)

    # PROTECTED_METHODS
    # -----------------

    def _import_publication_authors(self, publication: Dict[str, Any]):
        authors = publication['authors']
        authors = self._import_authors(authors)
        for author in authors:
            # Attempt to post the authorings
            data = {
                'author': author['slug'],
                'publication': publication['uuid']
            }
            self.authoring.post(data)

    def _import_authors(self, authors: Iterable[Dict[str, Any]]):
        result = []
        for author in authors:
            # The actual author
            response = self.author.post_or_get(author, scopus_id=author['scopus_id'])
            result.append(response)

        return result




import os
import shutil

from pathlib import Path
from typing import Iterable, Dict, Any

PATH = Path(__file__).parent.absolute()


# INSTALLATION UTILITIES
# ======================

def get_home_path() -> str:
    return str(Path.home())


def get_installation_path() -> str:
    # TODO: I think I would need to change this for windows and MAC?
    home_path = get_home_path()
    # Linux
    return os.path.join(home_path, '.pypubtrack')


def get_config_path() -> str:
    folder_path = get_installation_path()
    return os.path.join(folder_path, 'config.toml')


def check_installation() -> bool:
    installation_path = get_installation_path()
    return os.path.exists(installation_path) and os.path.isdir(installation_path)


def init_installation() -> str:
    # Create a new folder
    folder_path = get_installation_path()
    os.mkdir(folder_path, 0o777)

    # Copy the default config file into it
    config_template_path = os.path.join(PATH, 'templates', 'config.toml')
    config_path = get_config_path()
    shutil.copyfile(config_template_path, config_path)

    # Copy the version file into it
    version_template_path = os.path.join(PATH, 'VERSION')
    version_path = os.path.join(folder_path, 'VERSION')
    shutil.copyfile(version_template_path, version_path)

    return folder_path


# DICT UTILITIES
# ==============

def exclude_keys(d: Dict[Any, Any], keys: Iterable[Any]) -> Dict[Any, Any]:
    copy = d.copy()
    for key in keys:
        del copy[key]

    return copy


# MISC. UTILITIES
# ===============

def get_version() -> str:
    version_path = os.path.join(PATH, 'VERSION')
    with open(version_path, mode='r') as file:
        return file.read().replace('\n', '').replace(' ', '')


def author_name_kitopen(first_name: str, last_name: str) -> str:
    return '{}, {}*'.format(
        last_name.upper(),
        first_name[0].upper()
    )





import os
import toml

import pypubtrack.authentication as authentication

# CONFIG CLASS
# ============


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):
    """
    This is a singleton class, which implements the access to the config file.

    **Design Choice**

    So I feel like I want to explain my reasoning for this class here. This is indeed not my first implementation for
    config access within this project. Previously I was simply loading the dictionary into global variable within this
    file and then I could import this global variable and just use the dictionary.

    But while working with that implementation I have found two major problems with it:

    (1) Reloading: The problem is that the dictionary is loaded into the global variable at exactly the point when
    it is imported the first time at some other module. This means that there is no way of globally reloading the
    config to react to file system changes for example. Even if you were to assign the variable with a new dict
    for example, this change would not translate to other files! The reloaded version would only be present in a single
    file. With a singleton of course you can simply solve this problem with a "reload" method.

    (2) Reacting to change: There is a significant downside to using the config dict directly without some sort of
    intermediary layer. If you need some functionality say "CONFIG['camera']['sensor_height']" then this is the path
    which is directly implemented like this in the config file structure. This means that if you were to change the
    structure of the config file you would have to change every occasion in the code, which uses this attribute...
    This is obviously a bad SOC. With a class you could write methods, which wrap certain behaviour. After a change
    only this method would have to be changed.
    """

    def __init__(self):
        self.data = {}

    # IMPLEMENTING DICT FUNCTIONALITY
    # -------------------------------

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data.keys()

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    # WRAPPER METHODS
    # ---------------

    def get_url(self) -> str:
        return self.data['basic']['url']

    def get_token(self) -> str:
        return self.data['auth']['token']

    def get_authentication_class(self) -> type:
        class_string = self.data['auth']['type']
        return getattr(authentication, class_string)

    # HELPER FUNCTIONS
    # ----------------

    def load_dict(self, data: dict):
        self.data = data

    def load_file(self, file_path: str):
        data = toml.load(file_path)
        self.load_dict(data)




import os
import sys
import inspect
from typing import Dict, Any, Type, Union

import requests

from pypubtrack.config import Config


class Endpoint:
    """
    This is the base class to define a endpoint for a rest api.

    **Background**

    An "endpoint" refers to the URL, which is used to access some specific record/asset from the REST API. The
    pubtrack API exposes multiple of these endpoints to access/modify the internal database models of the app.

    **Details**

    A REST API is built on the basic HTTP protocol. All resources are in some way accessed by using these HTTP methods.
    This class implements the HTTP methods DELETE, PUT, PATCH, POST, GET to be sent to the specific endpoint, which
    is described by it. On top of that it also defines some composite methods, which combine these methods.

    This is only an abstract base class however. Every subclass has to implement the "get_endpoint" function, which
    is meant to return the concrete string URL for the specific endpoint it describes. The actual functionality of
    interaction however is universal and thus implemented in this very base class.

    **Example**

    Assume the following example of how to create an Endpoint subclass and how to for example request all of its
    results.

    .. code-block:: python

        import os

        class BooksEndpoint(Endpoint):

            def __init__(self, url, config):
                Endpoint.__init__(self, url, config)

            def get_endpoint():
                return 'books'

        # Describes the endpoint "http://myurl.com/api/books"
        books_endpoint = BooksEndpoint("http://myurl.com/api")
        books = books_endpoint.get()
        for book in books['results']:
            print(book)

    *A note on primary keys*

    Most of the methods require a primary key of the model to uniquely identify the database record on which to perform
    the action on. In terms of the previous example a "get" method with no parameters at all would return the list of
    all available books. But given a primary key, which in this case could be the string name, the method will only
    return the single record which is identified by this key.

    .. code-block:: python

        specific_book = books_endpoint.get('ultralearning')
        print(specific_book['title'])

    Now, in database modeling there is the concept of many-to-many relationships, which essentially can be illustrated
    as the "citation" relationship in the book example. A book can cite many other books, but it can also be cited by
    many other books. This kind of relationship is usually modeled as an intermediary database table called "citations"
    which has essentially two primary keys that uniquely identify the relationship. The first being the title of the
    book which cites and the second the title of the book which is cited. This kind of table would then be mapped to
    a REST API as having two primary keys "http://myurl.com/api/citation/title1/title2". This kind of resource can be
    accessed by simply passing two arguments as primary keys.

    .. code-block:: python

        specific_citation = citation_endpoint.get('ultralearning', 'thinking-fast-thinking-slow')

    """
    def __init__(self, url: str, config: Config):
        self.url = os.path.join(url, self.get_endpoint())
        self.config = config

    # BASIC API OPERATIONS
    # --------------------
    # These operations correlate directly with the basic HTTP methods with the same name.

    def delete(self, *pk: str):
        """
        Sends a DELETE request for the given primary key *pk*.
        A delete request will attempt to delete the record, which is identified by the primary key.

        :param pk:
        :return:
        """
        return self._request('delete', {
            'url':          self._get_url(*pk)
        })

    def put(self, *pk, data: dict = {}):
        """
        Sends a PUT request for the given primary key *pk* and the *data* dict.
        A put request is used to modify an already existing record which is identified by the given primary key. This
        method is essentially an "overwrite" in the sense that the existing record gets overwritten by the given data
        dict. Thus the data dict has to represent all necessary data which would have been necessary for a POST
        operation.

        :param pk:
        :param data:
        :return:
        """
        return self._request('put', {
            'url':          self._get_url(*pk),
            'json':         data
        })

    def patch(self, *pk, patch: dict = {}):
        """
        Sends a PATCH request for the given primary key *pk* and the *patch* dict.

        A patch request is used to modify an already existing record which is identified by the given primary key.
        Contrary to the put operation, the patch dict only has to contain those very attributes which are actually
        supposed to be changed.

        :param pk:
        :param patch:
        :return:
        """
        return self._request('patch', {
            'url':          self._get_url(*pk),
            'json':         patch
        })

    def post(self, data: dict = {}):
        """
        Sends a POST request for the given *data* dict.
        A post request is used to insert a new record. The data dict has to contain all the necessary information to
        construct a new record within the database of the application.

        :param data:
        :return:
        """
        return self._request('post', {
            'url':          self._get_url(),
            'json':         data
        })

    def get(self, *pk, params: dict = {}):
        """
        Sends a GET request for the given primary key *pk* and the additional get url parameters *params*.
        A get request is used to read the information of a specific record identified by the given primary key.

        For this method the primary key is optional. Calling this method without any primary keys given will return a
        list of ALL the records for that endpoint.

        :param pk:
        :param params:
        :return:
        """
        return self._request('get', {
            'url':          self._get_url(*pk),
            'params':       params
        })

    # COMPOSITE API OPERATIONS
    # ------------------------
    # These methods do not have a basic HTTP equivalent, but instead are utility functions, which are a useful
    # combination of a series of basic http methods.

    def post_or_get(self, data: dict, **get):
        """
        This method first attempts to post a new record with the given *data* dict. If this insert fails because a
        record with the given unique keys already exists, it will instead fetch this data. In any case after executing
        this method, one can be sure that a record with the given unique properties exists in the application and the
        details are returned.

        :param data:
        :param get:
        :return:
        """
        try:
            return self.post(data)
        except ConnectionError:
            return self.get_by(**get)

    def get_by(self, **params):
        """
        Retrieve a record by unique properties alternative to the primary key.

        This function simply calls the "get" method with the given *params* dict for the url parameters. The idea is
        that the REST api implements some sort of search mechanism through the url parameters and in this case
        records can be searched by secondary properties.

        :param params:
        :return:
        """
        results = self.get('', params=params)['results']
        if len(results) == 0:
            raise FileNotFoundError('{} not found {}'.format(str(self), str(params)))
        return results[0]

    # PROTECTED METHODS
    # -----------------

    def _request(self, method: str, kwargs: dict):
        kwargs = self._authentication(kwargs)

        func = getattr(requests, method)
        response = func(**kwargs)

        if response.status_code in [400]:
            raise ConnectionError('Request "{}" with status code: {} (kwargs: {})'.format(
                method,
                response.status_code,
                str(kwargs))
            )
        else:
            return response.json()

    def _get_url(self, *args):
        url = os.path.join(self.url, *args) if args else self.url
        if url[-1] != '/':
            url += '/'
        return url

    def _authentication(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        authentication_class = self.config.get_authentication_class()
        authenticate = authentication_class(self.config)
        return authenticate(kwargs)

    # MAGIC METHODS
    # -------------

    def __call__(self, pk=''):
        """
        An Endpoint object can also be called directly without any method. This is a convenience feature. This call
        will return the same as calling the "get" method without primary keys.

        **Example**

        .. code-block:: python

            # Both calls do the same thing
            books_endpoint.get()
            books_endpoint()

        :param pk:
        :return:
        """
        self.get(pk)

    def __str__(self):
        """
        Returns the string representation of a Endpoint object, which is simply a formatted version of the URL, which
        is returned by the "get_endpoint" method.
        :return:
        """
        return "ENDPOINT {}".format(self.url)

    # ABSTRACT METHODS
    # ----------------

    def get_endpoint(self):
        """
        This method has to be implemented by sub class!

        This method has to return the RELATIVE path towards the endpoint which is supposed to be described.
        Each endpoint object is created with a base URL which is the URL of the web applications api backend usually.
        This method is supposed to return the string of what substring has to be added to the end of this URL to
        arrive at the correct endpoint.

        :return:
        """
        raise NotImplementedError()


class AddEndpoint:
    """
    This is the class for a class decorator.

    This class can be used to add endpoints to a class as properties. The functionality is best described by an example:

    .. code-block:: python

        class BooksEndpoint(Endpoint):

            def __init__(self, url, config):
                Endpoint.__init__(self, url, config)

            def get_endpoint(self):
                return 'books'

        @AddEndpoint('books', BooksEndpoint)
        class Base:

            def __init__(self):
                self.url = 'http://myurl.com/api'


        base = Base()
        print(base.books) # "ENDPOINT http://myurl.com/api/books"
        books = base.books.get()['results']

    So, the decorator requires to arguments: A string name and a Enpoint class. It will then dynamically add an endpoint
    property with the given name to the decorated class, which is an instance of the given endpoint class. The only
    requirement which has to be fulfilled by the base class which is decorated is that it does not already have a
    method/attribute with the given name and *that it has a "self.url" attribute* which will be used as the basis for
    the endpoint!

    *Endpoint chaining*

    tbd

    """
    def __init__(self, name: str, endpoint: Union[Endpoint, str]):
        self.name = name
        self.endpoint = endpoint

    def __call__(self, cls: Type):

        def anonymous(this):
            if isinstance(self.endpoint, str):
                endpoint_class = self._lazy_class(self.endpoint, this.__module__)
                return endpoint_class(this.url, this.config)
            else:
                return self.endpoint(this.url, this.config)

        anonymous.__name__ = self.name
        setattr(cls, self.name, property(anonymous))

        return cls

    @classmethod
    def _lazy_class(cls, class_name: str, module: str):
        members = inspect.getmembers(sys.modules[module])
        members_dict = dict(members)
        return members_dict[class_name]
